function [V, S, max_res] = rat_Krylov_GenLyap( A, N, B, m, smin, smax )
%RAT_KRYLOV_GENLYAP
%  Using a naive block-Arnoldi to compute the rational Krylov type subspace
%  aimed at approximating the solution to the generalized Lyapunov equation
%             AX + XA^T + \sum_{i=1}^m N{i}XN{i}^T + BB^T = 0.
%  
%  The space created is
%      V = {B, (A-S(1))^{-1}B, (A-S(1))^{-1}(A-S(2))^{-1}B, ... , (A-S(1))^{-1}...(A-S(k))^{-1}B},
%  Shifts S are computed in a greedy fashin that minimize the current direction
%  Returns a space of dimension size(B,2)+(m-1)*size(B,2), that is, it does m-1 steps
%
%  [V, S, max_res] = rat_Krylov_GenLyap( A, N, B, m, smin, smax )
%
%  Input:
%  A - Matrix defining the Lyapunov operator
%  N - Cell array containing the "generalized" terms of the equation
%  B - Low-rank factor of the "right-hand-side"
%  m - The method does m-1 steps
%  smin - Estimate of the smallest eigenvalue of A
%  smax - Estimate of the largest eigenvalue of A
%
%  Output:
%  V - A matrix containing an orthogonal basis of the suggested search
%    space for approximating the generalized Lyapunov equation.
%  s - The used shifts
%  max_res - The maximum residuals at where the corresponding shifts were found
%

if ~iscell(N); error('N needs to be a cell-array'); end;

if smax < smin
  error('Input error: smax < smin')
end
if (abs(smin) < 1e-12 && smax == 0) || (abs(smax-smin)/smax < 1e-12)
  s_range = smax;
else
  s_range = linspace(smin, smax, 30);
end

n = size(A,1);
mm = m-1;

% V = zeros(n, r*m);
S = NaN((mm), 1);
max_res = -inf*ones((mm), 1);

if issparse(A)
  I = speye(n,n);
else
  I = eye(n,n);
end



V = qr_orth(B); %Orthogonalize and normalize within starting block
rb = size(V,2);

for k = 1:(mm)
  fprintf('Iteration: %2.i of %i\n',k,(mm))
  BB = full(V(:,((k-1)*rb+1):k*rb));

  % Find a new shift
  % Search the grid for a good point. New point where residual is maximum.
  nn = size(V,2);
  II = speye(nn, nn);
  s = NaN;
  Ap = V'*A*V;
  Bp = V'*BB;
  for ss = s_range
    %       res = norm(R - (ss*I - A) * (V * ((ss*II - Ap)\Rp)));
    % Below is an optimized way. Make sure to work with vectors!
    r1 = (V * ((ss*II - Ap)\Bp));
    r3 = A*r1;
    r4 = ss*r1;
    res = norm(BB-r4+r3);
    if(res > max_res(k))
      max_res(k) = res;
      s = ss;
    end
  end
  S(k) = s;
 
  % Expand the basis
  Vi = (A - s*I)\BB;
  V = qr_orth([V,Vi]);  
end

end